use crate::{
    bundle::{BundleType, BundleValue, DynBundle},
    firrtl,
    intern::Interned,
    module::Module,
};
use clap::{
    builder::{OsStringValueParser, TypedValueParser},
    Args, Parser, Subcommand, ValueEnum, ValueHint,
};
use eyre::{eyre, Report};
use std::{error, ffi::OsString, fmt, io, path::PathBuf, process};

pub type Result<T = (), E = CliError> = std::result::Result<T, E>;

pub struct CliError(Report);

impl fmt::Debug for CliError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

impl fmt::Display for CliError {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        self.0.fmt(f)
    }
}

impl error::Error for CliError {}

impl From<io::Error> for CliError {
    fn from(value: io::Error) -> Self {
        CliError(Report::new(value))
    }
}

pub trait RunPhase<Arg> {
    type Output;
    fn run(&self, arg: Arg) -> Result<Self::Output>;
}

#[derive(Args, Debug)]
#[non_exhaustive]
pub struct BaseArgs {
    /// the directory to put the generated main output file and associated files in
    #[arg(short, long, value_hint = ValueHint::DirPath)]
    pub output: PathBuf,
    /// the stem of the generated main output file, e.g. to get foo.v, pass --file-stem=foo
    #[arg(long)]
    pub file_stem: Option<String>,
}

impl BaseArgs {
    pub fn to_firrtl_file_backend(&self) -> firrtl::FileBackend {
        firrtl::FileBackend {
            dir_path: self.output.clone(),
            top_fir_file_stem: self.file_stem.clone(),
        }
    }
}

#[derive(Args, Debug)]
#[non_exhaustive]
pub struct FirrtlArgs {
    #[command(flatten)]
    pub base: BaseArgs,
}

#[derive(Debug)]
#[non_exhaustive]
pub struct FirrtlOutput {
    pub file_stem: String,
}

impl FirrtlOutput {
    pub fn firrtl_file(&self, args: &FirrtlArgs) -> PathBuf {
        let mut retval = args.base.output.join(&self.file_stem);
        retval.set_extension("fir");
        retval
    }
}

impl FirrtlArgs {
    fn run_impl(&self, top_module: Module<DynBundle>) -> Result<FirrtlOutput> {
        let firrtl::FileBackend {
            top_fir_file_stem, ..
        } = firrtl::export(self.base.to_firrtl_file_backend(), &top_module)?;
        Ok(FirrtlOutput {
            file_stem: top_fir_file_stem.expect(
                "export is known to set the file stem from the circuit name if not provided",
            ),
        })
    }
}

impl<T: BundleValue> RunPhase<Module<T>> for FirrtlArgs
where
    T::Type: BundleType<Value = T>,
{
    type Output = FirrtlOutput;
    fn run(&self, top_module: Module<T>) -> Result<Self::Output> {
        self.run_impl(top_module.canonical())
    }
}

impl<T: BundleValue> RunPhase<Interned<Module<T>>> for FirrtlArgs
where
    T::Type: BundleType<Value = T>,
{
    type Output = FirrtlOutput;
    fn run(&self, top_module: Interned<Module<T>>) -> Result<Self::Output> {
        self.run(*top_module)
    }
}

/// based on [LLVM Circt's recommended lowering options
/// ](https://circt.llvm.org/docs/VerilogGeneration/#recommended-loweringoptions-by-target)
#[derive(ValueEnum, Copy, Clone, Debug, PartialEq, Eq, Hash)]
#[non_exhaustive]
pub enum VerilogDialect {
    Questa,
    Spyglass,
    Verilator,
    Vivado,
    Yosys,
}

impl VerilogDialect {
    pub fn firtool_extra_args(self) -> &'static [&'static str] {
        match self {
            VerilogDialect::Questa => &["--lowering-options=emitWireInPorts"],
            VerilogDialect::Spyglass => {
                &["--lowering-options=explicitBitcast,disallowExpressionInliningInPorts"]
            }
            VerilogDialect::Verilator => &[
                "--lowering-options=locationInfoStyle=wrapInAtSquareBracket,disallowLocalVariables",
            ],
            VerilogDialect::Vivado => &["--lowering-options=mitigateVivadoArrayIndexConstPropBug"],
            VerilogDialect::Yosys => {
                &["--lowering-options=disallowLocalVariables,disallowPackedArrays"]
            }
        }
    }
}

#[derive(Args, Debug)]
#[non_exhaustive]
pub struct VerilogArgs {
    #[command(flatten)]
    pub firrtl: FirrtlArgs,
    #[arg(
        long,
        default_value = "firtool",
        env = "FIRTOOL",
        value_hint = ValueHint::CommandName,
        value_parser = OsStringValueParser::new().try_map(which::which)
    )]
    pub firtool: PathBuf,
    #[arg(long)]
    pub firtool_extra_args: Vec<OsString>,
    /// adapt the generated Verilog for a particular toolchain
    #[arg(long)]
    pub verilog_dialect: Option<VerilogDialect>,
}

#[derive(Debug)]
#[non_exhaustive]
pub struct VerilogOutput {
    pub firrtl: FirrtlOutput,
}

impl VerilogOutput {
    pub fn verilog_file(&self, args: &VerilogArgs) -> PathBuf {
        let mut retval = args.firrtl.base.output.join(&self.firrtl.file_stem);
        retval.set_extension("v");
        retval
    }
}

impl VerilogArgs {
    fn run_impl(&self, firrtl_output: FirrtlOutput) -> Result<VerilogOutput> {
        let output = VerilogOutput {
            firrtl: firrtl_output,
        };
        let mut cmd = process::Command::new(&self.firtool);
        cmd.arg(output.firrtl.firrtl_file(&self.firrtl));
        cmd.arg("-o");
        cmd.arg(output.verilog_file(self));
        if let Some(dialect) = self.verilog_dialect {
            cmd.args(dialect.firtool_extra_args());
        }
        cmd.args(&self.firtool_extra_args);
        cmd.current_dir(&self.firrtl.base.output);
        let status = cmd.status()?;
        if status.success() {
            Ok(output)
        } else {
            Err(CliError(eyre!(
                "running {} failed: {status}",
                self.firtool.display()
            )))
        }
    }
}

impl<Arg> RunPhase<Arg> for VerilogArgs
where
    FirrtlArgs: RunPhase<Arg, Output = FirrtlOutput>,
{
    type Output = VerilogOutput;
    fn run(&self, arg: Arg) -> Result<Self::Output> {
        let firrtl_output = self.firrtl.run(arg)?;
        self.run_impl(firrtl_output)
    }
}

#[derive(Subcommand, Debug)]
enum CliCommand {
    /// Generate FIRRTL
    Firrtl(FirrtlArgs),
    /// Generate Verilog
    Verilog(VerilogArgs),
}

/// a simple CLI
///
/// Use like:
///
/// ```no_run
/// # use fayalite::hdl_module;
/// # #[hdl_module]
/// # fn my_module() {}
/// use fayalite::cli;
///
/// fn main() -> cli::Result {
///     cli::Cli::parse().run(my_module())
/// }
/// ```
///
/// You can also use it with a larger [`clap`]-based CLI like so:
///
/// ```no_run
/// # use fayalite::hdl_module;
/// # #[hdl_module]
/// # fn my_module() {}
/// use clap::{Subcommand, Parser};
/// use fayalite::cli;
///
/// #[derive(Subcommand)]
/// pub enum Cmd {
///     #[command(flatten)]
///     Fayalite(cli::Cli),
///     MySpecialCommand {
///         #[arg(long)]
///         foo: bool,
///     },
/// }
///
/// #[derive(Parser)]
/// pub struct Cli {
///     #[command(subcommand)]
///     cmd: Cmd, // or just use cli::Cli directly if you don't need more subcommands
/// }
///
/// fn main() -> cli::Result {
///     match Cli::parse().cmd {
///         Cmd::Fayalite(v) => v.run(my_module())?,
///         Cmd::MySpecialCommand { foo } => println!("special: foo={foo}"),
///     }
///     Ok(())
/// }
/// ```
#[derive(Parser, Debug)]
// clear things that would be crate-specific
#[command(name = "Fayalite Simple CLI", about = None, long_about = None)]
pub struct Cli {
    #[command(subcommand)]
    subcommand: CliCommand,
}

impl clap::Subcommand for Cli {
    fn augment_subcommands(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands(cmd)
    }

    fn augment_subcommands_for_update(cmd: clap::Command) -> clap::Command {
        CliCommand::augment_subcommands_for_update(cmd)
    }

    fn has_subcommand(name: &str) -> bool {
        CliCommand::has_subcommand(name)
    }
}

impl<T> RunPhase<T> for Cli
where
    FirrtlArgs: RunPhase<T, Output = FirrtlOutput>,
{
    type Output = ();
    fn run(&self, arg: T) -> Result<Self::Output> {
        match &self.subcommand {
            CliCommand::Firrtl(c) => {
                c.run(arg)?;
            }
            CliCommand::Verilog(c) => {
                c.run(arg)?;
            }
        }
        Ok(())
    }
}

impl Cli {
    /// forwards to [`clap::Parser::parse()`] so you don't have to import [`clap::Parser`]
    pub fn parse() -> Self {
        clap::Parser::parse()
    }
    /// forwards to [`RunPhase::run()`] so you don't have to import [`RunPhase`]
    pub fn run<T>(&self, top_module: T) -> Result<()>
    where
        Self: RunPhase<T, Output = ()>,
    {
        RunPhase::run(self, top_module)
    }
}
